#include <floppy.h>
#include <io.h>
#include <process.h>
#include <fat12.h>
/* globals */
//static IrqSave oldirq6;
//static IrqSave oldint1c;
static volatile BOOL done = FALSE;
static BOOL dchange = FALSE;
static BOOL motor = FALSE;
static int mtick = 0;
static volatile int tmout = 0;
static BYTE status[7] = { 0 };
static BYTE statsz = 0;
static BYTE sr0 = 0;
static BYTE fdc_track = 0xff;
static DrvGeom geometry = { DG144_HEADS,DG144_TRACKS,DG144_SPT };

static long tbaddr;    /* physical address of track buffer located below 1M */

/* prototypes */
void sendbyte(int byte);
int getbyte();
void irq6(void);
void int1c(void);
BOOL waitfdc(BOOL sensei);
BOOL fdc_rw(int block,BYTE *blockbuff,BOOL read);

/* helper functions */

void floppy_interrupt(void);

/* init driver */
void init_floppy(void)
{
   int i;
   
   /* allocate track buffer (must be located below 1M) */
   tbaddr = alloc_dma_buffer(512);
   
   /* install IRQ6 handler */
   //set_irq_handler(6,irq6,&oldirq6);
   
   add_idt(0x26,floppy_interrupt);
   outb(inb_p(0x21)&~0x40,0x21);
   printk("IRQ6 handler installed\n");

//   set_irq_handler(0x1c,int1c,&oldint1c);
//   printf("timer handler installed\n");

   reset();

   /* get floppy controller version */
   sendbyte(CMD_VERSION);
   i = getbyte();

   if (i == 0x80)
     printk("NEC765 controller found\n");
   else
     printk("enhanced controller found\n");
   log_disk(NULL);
}

/* sendbyte() routine from intel manual */
void sendbyte(int byte)
{
   volatile int msr;
   int tmo;
   
   for (tmo = 0;tmo < 128;tmo++) {
      msr = inb(FDC_MSR);
      if ((msr & 0xc0) == 0x80) {
	 outb_p(byte,FDC_DATA);
	 return;
      }
      inb(0x80);   /* delay */
   }
}

/* getbyte() routine from intel manual */
int getbyte()
{
   volatile int msr;
   int tmo;
   
   for (tmo = 0;tmo < 128;tmo++) {
      msr = inb(FDC_MSR);
      if ((msr & 0xd0) == 0xd0) {
	 return inb(FDC_DATA);
      }
      inb(0x80);   /* delay */
   }

   return -1;   /* read timeout */
}

/* this waits for FDC command to complete */
BOOL waitfdc(BOOL sensei)
{
   tmout = 18;   /* set timeout to 1 second */
     
   /* wait for IRQ6 handler to signal command finished */
   while (!done && tmout)
     ;
   
   /* read in command result bytes */
   statsz = 0;
   while ((statsz < 7) && (inb(FDC_MSR) & (1<<4))) {
      status[statsz++] = getbyte();
   }

   if (sensei) {
      /* send a "sense interrupt status" command */
      sendbyte(CMD_SENSEI);
      sr0 = getbyte();
      fdc_track = getbyte();
   }
   
   done = FALSE;
   
   if (!tmout) {
      /* timed out! */
      if (inb(FDC_DIR) & 0x80)  /* check for diskchange */
	dchange = TRUE;
      
      return FALSE;
   } else
     return TRUE;
}

/* This is the IRQ6 handler */
void irq6(void)
{
   /* signal operation finished */
   done = TRUE;
//    printk("IRQ6!!!!");
   /* EOI the PIC */
   outb_p(0x20,0x20);
}

/* This is the timer (int 1ch) handler */
void int1c(void)
{
   if (tmout) --tmout;     /* bump timeout */
   
   if (mtick > 0)
     --mtick;
   else if (!mtick && motor) {
      outb_p(0x0c,FDC_DOR);  /* turn off floppy motor */
      motor = FALSE;
   }
}

/*
 * converts linear block address to head/track/sector
 * 
 * blocks are numbered 0..heads*tracks*spt-1
 * blocks 0..spt-1 are serviced by head #0
 * blocks spt..spt*2-1 are serviced by head 1
 * 
 * WARNING: garbage in == garbage out
 */
void block2hts(int block,int *head,int *track,int *sector)
{
   *head = (block % (geometry.spt * geometry.heads)) / (geometry.spt);
   *track = block / (geometry.spt * geometry.heads);
   *sector = block % geometry.spt + 1;
}

/**** disk operations ****/

/* this gets the FDC to a known state */
void reset(void)
{
   /* stop the motor and disable IRQ/DMA */
   outb_p(0,FDC_DOR);
   
   mtick = 0;
   motor = FALSE;

   /* program data rate (500K/s) */
   outb_p(0,FDC_DRS);

   /* re-enable interrupts */
   outb_p(0x0c,FDC_DOR);

   /* resetting triggered an interrupt - handle it */
   done = TRUE;
   waitfdc(TRUE);

   /* specify drive timings (got these off the BIOS) */
   sendbyte(CMD_SPECIFY);
   sendbyte(0xdf);  /* SRT = 3ms, HUT = 240ms */
   sendbyte(0x02);  /* HLT = 16ms, ND = 0 */
   
   /* clear "disk change" status */
   seek(1);
   recalibrate();

   dchange = FALSE;
}

/* this returns whether there was a disk change */
BOOL diskchange(void)
{
   return dchange;
}

/* this turns the motor on */
void motoron(void)
{
   if (!motor) {
      mtick = -1;     /* stop motor kill countdown */
      outb_p(0x1c,FDC_DOR);
  //    printk("avant ksleep\n");
      k_msleep(500); /* delay 500ms for motor to spin up */
  //    printk("apres ksleep\n");
      motor = TRUE;
   }
}

/* this turns the motor off */
void motoroff(void)
{
   if (motor) {
      mtick = 36;   /* start motor kill countdown: 36 ticks ~ 2s */
   }
}

/* recalibrate the drive */
void recalibrate(void)
{
   /* turn the motor on */
   motoron();
   
   /* send actual command bytes */
   sendbyte(CMD_RECAL);
   sendbyte(0);

   /* wait until seek finished */
   waitfdc(TRUE);
   
   /* turn the motor off */
   motoroff();
}

/* seek to track */
BOOL seek(int track)
{
   if (fdc_track == track)  /* already there? */
     return TRUE;
 
   motoron();
   
   /* send actual command bytes */
   sendbyte(CMD_SEEK);
   sendbyte(0);
   sendbyte(track);

   /* wait until seek finished */
   if (!waitfdc(TRUE))
     return FALSE;     /* timeout! */

   /* now let head settle for 15ms */
   k_msleep(15);
   
   motoroff();
   
   /* check that seek worked */
   if ((sr0 != 0x20) || (fdc_track != track))
     return FALSE;
   else
     return TRUE;
}

/* checks drive geometry - call this after any disk change */
BOOL log_disk(DrvGeom *g)
{
   /* get drive in a known status before we do anything */
   reset();

   /* assume disk is 1.68M and try and read block #21 on first track */
   geometry.heads = DG168_HEADS;
   geometry.tracks = DG168_TRACKS;
   geometry.spt = DG168_SPT;

   if (read_block(20,NULL)) {
      /* disk is a 1.68M disk */
      if (g) {
	 g->heads = geometry.heads;
	 g->tracks = geometry.tracks;
	 g->spt = geometry.spt;
      }
      return TRUE;             
   }
   
   /* OK, not 1.68M - try again for 1.44M reading block #18 on first track */
   geometry.heads = DG144_HEADS;
   geometry.tracks = DG144_TRACKS;
   geometry.spt = DG144_SPT;

   if (read_block(17,NULL)) {
      /* disk is a 1.44M disk */
      if (g) {
	 g->heads = geometry.heads;
	 g->tracks = geometry.tracks;
	 g->spt = geometry.spt;
      }
      return TRUE;
   }
   
   /* it's not 1.44M or 1.68M - we don't support it */
   return FALSE;
}

/* read block (blockbuff is 512 byte buffer) */
BOOL read_block(int block,BYTE *blockbuff)
{
   return fdc_rw(block,blockbuff,TRUE);
}

/* write block (blockbuff is 512 byte buffer) */
BOOL write_block(int block,BYTE *blockbuff)
{
   return fdc_rw(block,blockbuff,FALSE);
}

/*
 * since reads and writes differ only by a few lines, this handles both.  This
 * function is called by read_block() and write_block()
 */
BOOL fdc_rw(int block,BYTE *blockbuff,BOOL read)
{
   int head,track,sector,tries,i;

   /* convert logical address into physical address */
   block2hts(block,&head,&track,&sector);
//   printf("block %d = %d:%02d:%02d\n",block,head,track,sector);
   
   /* spin up the disk */
   motoron();

   if (!read && blockbuff) {
      /* copy data from data buffer into track buffer */
      memcpy((void*)tbaddr,blockbuff,512);
   }
   
   for (tries = 0;tries < 3;tries++) {
      /* check for diskchange */
      if (inb(FDC_DIR) & 0x80) {
	 dchange = TRUE;
	 seek(1);  /* clear "disk change" status */
	 recalibrate();
	 motoroff();
	 return FALSE;
      }
         
      /* move head to right track */
      if (!seek(track)) {
	 motoroff();
	 return FALSE;
      }
      
      /* program data rate (500K/s) */
      outb_p(0,FDC_CCR);
      
      /* send command */
      if (read) {
	 dma_xfer(2,tbaddr,512,FALSE);
	 sendbyte(CMD_READ);
      } else {
	 dma_xfer(2,tbaddr,512,TRUE);
	 sendbyte(CMD_WRITE);
      }
      
      sendbyte(head << 2);
      sendbyte(track);
      sendbyte(head);
      sendbyte(sector);
      sendbyte(2);               /* 512 bytes/sector */
      sendbyte(geometry.spt);
      if (geometry.spt == DG144_SPT)
	sendbyte(DG144_GAP3RW);  /* gap 3 size for 1.44M read/write */
      else
	sendbyte(DG168_GAP3RW);  /* gap 3 size for 1.68M read/write */
      sendbyte(0xff);            /* DTL = unused */
      
      /* wait for command completion */
      /* read/write don't need "sense interrupt status" */
      if (!waitfdc(FALSE))
	return FALSE;   /* timed out! */
      
      if ((status[0] & 0xc0) == 0) break;   /* worked! outta here! */
   
      recalibrate();  /* oops, try again... */
   }
   
   /* stop the motor */
   motoroff();

   if (read && blockbuff) {
      /* copy data from track buffer into data buffer */
      memcpy(blockbuff,(void*)tbaddr,512);
   }

/*   printk("status bytes: ");
   for (i = 0;i < statsz;i++)
     printk("%02x ",status[i]);

    printk("\n");*/

   return (tries != 3);
}

/* this formats a track, given a certain geometry */
BOOL format_track(BYTE track,DrvGeom *g)
{
   int i,h,r,r_id,split;
   BYTE tmpbuff[256];

   /* check geometry */
   if (g->spt != DG144_SPT && g->spt != DG168_SPT)
     return FALSE;
   
   /* spin up the disk */
   motoron();

   /* program data rate (500K/s) */
   outb_p(0,FDC_CCR);

   seek(track);  /* seek to track */

   /* precalc some constants for interleave calculation */
   split = g->spt / 2;
   if (g->spt & 1) split++;
   
   for (h = 0;h < g->heads;h++) {
      /* for each head... */
      
      /* check for diskchange */
      if (inb(FDC_DIR) & 0x80) {
	 dchange = TRUE;
	 seek(1);  /* clear "disk change" status */
	 recalibrate();
	 motoroff();
	 return FALSE;
      }

      i = 0;   /* reset buffer index */
      for (r = 0;r < g->spt;r++) {
	 /* for each sector... */

	 /* calculate 1:2 interleave (seems optimal in my system) */
	 r_id = r / 2 + 1;
	 if (r & 1) r_id += split;
	 
	 /* add some head skew (2 sectors should be enough) */
	 if (h & 1) {
	    r_id -= 2;
	    if (r_id < 1) r_id += g->spt;
	 }
      
	 /* add some track skew (1/2 a revolution) */
	 if (track & 1) {
	    r_id -= g->spt / 2;
	    if (r_id < 1) r_id += g->spt;
	 }
	 
	 /**** interleave now calculated - sector ID is stored in r_id ****/

	 /* fill in sector ID's */
	 tmpbuff[i++] = track;
	 tmpbuff[i++] = h;
	 tmpbuff[i++] = r_id;
	 tmpbuff[i++] = 2;
      }

      /* copy sector ID's to track buffer */
      memcpy((void*)tbaddr,tmpbuff,i);
      
      /* start dma xfer */
      dma_xfer(2,tbaddr,i,TRUE);
      
      /* prepare "format track" command */
      sendbyte(CMD_FORMAT);
      sendbyte(h << 2);
      sendbyte(2);
      sendbyte(g->spt);
      if (g->spt == DG144_SPT)      
	sendbyte(DG144_GAP3FMT);    /* gap3 size for 1.44M format */
      else
	sendbyte(DG168_GAP3FMT);    /* gap3 size for 1.68M format */
      sendbyte(0);     /* filler byte */
	 
      /* wait for command to finish */
      if (!waitfdc(FALSE))
	return FALSE;
      
      if (status[0] & 0xc0) {
	 motoroff();
	 return FALSE;
      }
   }
   
   motoroff();
   
   return TRUE;
}
/* definition of DMA channels */
const static DmaChannel dmainfo[] = {
   { 0x87, 0x00, 0x01 },
   { 0x83, 0x02, 0x03 },
   { 0x81, 0x04, 0x05 },
   { 0x82, 0x06, 0x07 }
};

/* 
 * this allocates a buffer in the < 1M range, maps it and returns the linear
 * address, also setting the physical in the integer pointed at
 */

char _dma_buffer[512];

long alloc_dma_buffer(int size)
{

    return((long)_dma_buffer);

}

/*
 * this sets up a DMA trasfer between a device and memory.  Pass the DMA
 * channel number (0..3), the physical address of the buffer and transfer
 * length.  If 'read' is TRUE, then transfer will be from memory to device,
 * else from the device to memory.
 */
void dma_xfer(int channel,long physaddr,int length,BOOL read)
{
   long page,offset;
   
   if(channel > 3) return;
   
   /* calculate dma page and offset */
   page = physaddr >> 16;
   offset = physaddr & 0xffff;
   length -= 1;  /* with dma, if you want k bytes, you ask for k - 1 */

   __asm__("cli");  /* disable irq's */
   
   /* set the mask bit for the channel */
   outb_p(channel | 4,0x0a);
   
   /* clear flipflop */
   outb_p(0,0x0c);
   
   /* set DMA mode (write+single+r/w) */
   outb_p((read ? 0x48 : 0x44) + channel,0x0b);
   
   /* set DMA page */
   outb_p(page,dmainfo[channel].page);
   
   /* set DMA offset */
   outb_p(offset & 0xff,dmainfo[channel].offset);  /* low byte */
   outb_p(offset >> 8,dmainfo[channel].offset);    /* high byte */
   
   /* set DMA length */
   outb_p(length & 0xff,dmainfo[channel].length);  /* low byte */
   outb_p(length >> 8,dmainfo[channel].length);    /* high byte */
   
   /* clear DMA mask bit */
   outb_p(channel,0x0a);
   
   __asm__("sti");  /* enable irq's */
}

/* ensure all mem is locked */
//int _crt0_startup_flags = _CRT0_FLAG_LOCK_MEMORY;

/* test functions */

int format_floppy()
{
    int i;
    DrvGeom geometry;
    log_disk(&geometry);
    for (i = 0;i < geometry.tracks;i++) { 
	if (!format_track(i,&geometry)) {
	    if (diskchange())
		printk("diskchange - abort!\n");
	    else
		printk("\nerror!\n");
	    return 1;
	}
	printk("formatted track %d\n",i);
    }
    return 0;
}

int read_floppy(char * buff,int sector,int nb)
{
    int i;
    DrvGeom geometry;
    for(i=0;i<nb;i++) read_block(sector+i,buff+512*i);
    motoroff();
    outb_p(0x0c,FDC_DOR);  /* turn off floppy motor */
    motor = FALSE;
		
}
    

int test_floppy(void)
{
   int block,i,c,track;
   BYTE trackbuff[512];
   DrvGeom geometry;

   for (i = 0;i < 512;i++) trackbuff[i] = 0;
   
   init_floppy();
   
   
   if (!log_disk(&geometry)) {
      printk("cannot read geometry!\n");
      while(1);	
   }

   if (geometry.spt == DG144_SPT)
     printk("1.44M format\n");
   else
     printk("1.68M format\n");
   
   read_block(1000,trackbuff);
   printk(trackbuff);
   
   return 0;
}

struct disquette disque_a;

int pid_gestion_disque;

void gestion_disque()
{
    while(1)
    {
        current->state=STOPPED;
        current->ttl = 0;
	__asm__("int $0x68");  // dors jusqu'a un appel
	switch (disque_a.ordre)
	{
	    case 0:
	    {
		disque_a.retour=mount(disque_a.fd);
		disque_a.p->state=SYSTEM;
		break;
	    }
	    case 1:
	    {
		int i,j;
		char bu[255];
		i=get_first_libre(disque_a.p);
		if(i=-1)
		{
		    disque_a.retour=-1;
		    disque_a.p->state=SYSTEM;
		    break;
		}
		get_cr3(disque_a.p->context->cr3);
		strncpy(bu,disque_a.buf,255);
		get_system_cr3();
		j=openf(bu,&(disque_a.p->fichier[i]->data));
		
		if(j<0)
		{
		    disque_a.retour=j;
		    disque_a.p->state=SYSTEM;
		    break;
		}
		disque_a.retour=i;
		((Vortex*)disque_a.p->fichier[i]->data)->ID=i;
		break;
	    }
	    default:
	    disque_a.retour=-2;
	    disque_a.p->state=SYSTEM;
	    break;
	}
    }
}

