/*
 *  Barbux
 *  Copyright (C) 2003 Wons Jean-Baptiste
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * floppy.c - floppy disk handling
 */

#include <btypes.h>

static uchar version;
static uchar cmos_type;
static uchar type;

static uchar motor_mask = 0;

static uint cur_op = 0;
static uint cur_drv;
static uint cur_head;
static uint cur_trk;

//#define FD_SEEK   0x02

/*
 * max no of floppies supported
 */
#define     NFD      (2)

#define     FD_BASEIO   0x3F0

#define     F_MASTER 	0x80
#define     F_DIR    	0x40
#define     F_CMDBUSY   0x10

#define     FD_DATA     0x05
#define     FD_STATUS   0x04

/* 
 * fdc commands 
 */
#define     FDC_VERSION    0x10
#define     FDC_INTS       0x08  /* SENSEI */
#define     FDC_RESET      0x04  /* SENSE */
#define     FDC_SEEK       0x0F

#define     FDC_VER_765A   0x80
#define     FDC_VER_765B   0x90

#define     FDTYPE_MASK 0x0F

#define     FDTYPE_NONE 0x00
#define     FDTYPE_360  0x01
#define     FDTYPE_1200 0x02
#define     FDTYPE_720  0x03
#define     FDTYPE_1440 0x04
#define     FDTYPE_2880 0x05

#define     ENABLE_INTS 0x0C

#define     FD_MOTOR 	0x02

#define     MOTOR_0     0x10
#define     MOTOR_1     0x20

#define     RTCSEL      0x70
#define     RTCDATA     0x71

#define     NVRAM_FD 	0x10

#define cli() __asm__("cli")
#define sti() __asm__("sti")

/* Bits of main status register */
#define STATUS_BUSYMASK 0x0F            /* drive busy mask */
#define STATUS_BUSY     0x10            /* FDC busy */
#define STATUS_DMA      0x20            /* 0- DMA mode */
#define STATUS_DIR      0x40            /* 0- cpu->fdc */
#define STATUS_READY    0x80            /* Data reg ready */

/* Bits of FD_ST0 */
#define ST0_DS          0x03            /* drive select mask */
#define ST0_HA          0x04            /* Head (Address) */
#define ST0_NR          0x08            /* Not Ready */
#define ST0_ECE         0x10            /* Equipment chech error */
#define ST0_SE          0x20            /* Seek end */
#define ST0_INTR        0xC0            /* Interrupt code mask */

/* Bits of FD_ST1 */
#define ST1_MAM         0x01            /* Missing Address Mark */
#define ST1_WP          0x02            /* Write Protect */
#define ST1_ND          0x04            /* No Data - unreadable */
#define ST1_OR          0x10            /* OverRun */
#define ST1_CRC         0x20            /* CRC error in data or addr */
#define ST1_EOC         0x80            /* End Of Cylinder */

/* Bits of FD_ST2 */
#define ST2_MAM         0x01            /* Missing Addess Mark (again) */
#define ST2_BC          0x02            /* Bad Cylinder */
#define ST2_SNS         0x04            /* Scan Not Satisfied */
#define ST2_SEH         0x08            /* Scan Equal Hit */
#define ST2_WC          0x10            /* Wrong Cylinder */
#define ST2_CRC         0x20            /* CRC error in data field */
#define ST2_CM          0x40            /* Control Mark = deleted */

/* Bits of FD_ST3 */
#define ST3_HA          0x04            /* Head (Address) */
#define ST3_TZ          0x10            /* Track Zero signal (1=track 0) */
#define ST3_WP          0x40            /* Write Protect */

/* Values for FD_COMMAND */
#define FD_RECALIBRATE          0x07    /* move to track 0 */
#define FD_SEEK                 0x0F    /* seek track */
#define FD_READ                 0xE6    /* read with MT, MFM, SKip deleted */
#define FD_WRITE                0xC5    /* write with MT, MFM */
#define FD_SENSEI               0x08    /* Sense Interrupt Status */
#define FD_SPECIFY              0x03    /* specify HUT etc */
#define FD_FORMAT               0x4D    /* format one track */
#define FD_VERSION              0x10    /* get version code */
#define FD_CONFIGURE            0x13    /* configure FIFO operation */
#define FD_PERPENDICULAR        0x12    /* perpendicular r/w mode */

/* DMA commands */
#define DMA_READ        0x46
#define DMA_WRITE       0x4A

/* FDC version return types */
#define FDC_TYPE_STD    0x80    /* normal 8272A clone FDC */
#define FDC_TYPE_82077  0x90    /* FIFO + perpendicular support */

/*
 *  * globals used by 'result()'
 *   */
#define MAX_REPLIES 7
static unsigned char reply_buffer[MAX_REPLIES];
#define ST0 (reply_buffer[0])
#define ST1 (reply_buffer[1])
#define ST2 (reply_buffer[2])
#define ST3 (reply_buffer[3])


uchar floppy_in()
{
   uint c;
   int r;

   c = 1000000;

   do {
      r = (in(FD_BASEIO+FD_STATUS) 
         & (F_MASTER | F_DIR | F_CMDBUSY));
      if( r == (F_MASTER | F_DIR | F_CMDBUSY ))
         break;
      c--;
   } while(c);
   
   if( c == 0 )
      return 0;

   return in(FD_BASEIO+FD_DATA);
}

void floppy_out(byte val)
{
   uint c;
   int r;

   c = 1000000;

   do {
      r = (in(FD_BASEIO+FD_STATUS) & (F_MASTER | F_DIR));
      if( r == F_MASTER)
         break;
      c--;
   } while(c);

   if( c == 0 ) {
      printk("floppy_out: timeout\n");
      return;
   }
   
   out(FD_BASEIO+FD_DATA, val);
}

void floppy_reset()
{
   /*
   out(0x03f2, 0x08);
   for(c = 0; c < 10000; c++);
   out(0x03f2, 0x0C);
   */

   int c;

   cli();

   out(FD_BASEIO+FD_MOTOR, 0);
   out(FD_BASEIO+FD_MOTOR, ENABLE_INTS);

   sti();

   for(c = 0; c < 10000; c++);
}

void floppy_start(int drive)
{
   motor_mask |= drive;
   if( drive == 0x00 ) {
      motor_mask |= MOTOR_0;
   } else {
      motor_mask |= MOTOR_1;
   }
   out(FD_BASEIO+FD_MOTOR, motor_mask);
}

void floppy_stop(int drive)
{
   if( drive == 0x00 ) {
      motor_mask &= ~MOTOR_0;
   } else {
      motor_mask &= ~MOTOR_1;
   }
   out(FD_BASEIO+FD_MOTOR, motor_mask);
}

void floppy_seek()
{
   cur_op = FD_SEEK;
   cur_head = 0;
   cur_trk = 0;
   cur_drv = 0;

   floppy_out(FDC_SEEK);
   floppy_out((cur_head << 2) | cur_drv);
   floppy_out(cur_trk);
}


void init_floppy()
{
   int c;
   uint l;
   printk("Initialisation du floppy: ");	
   floppy_reset();

   floppy_out(FDC_VERSION);
   version = floppy_in();
   printk("Version %d trouvee: ", version);

   if( version == FDC_VER_765A )
      printk("stardard FDC trouve\n");
   else
   if( version == FDC_VER_765B )
      printk("enhanced FDC trouve\n");
   else
   if( version == 0 )
      printk("timeout\n");
   else
      printk("controlleur inconnu\n");

   for(c = 0; c < NFD; ++c) {
      out(RTCSEL, NVRAM_FD);
      cmos_type = in(RTCDATA);

      type = (cmos_type >> (4 * (NFD - 1 - c ))) & FDTYPE_MASK;

      switch(type) {
         case FDTYPE_1200:
            printk("floppy_init -> %d : disque 1.2MB trouve\n",c);
            break;
         case FDTYPE_1440:
            printk("floppy_init -> %d : disque 1.44MB trouve\n",c);
            break;
         default:
            printk("floppy_init -> %d : pas de disque\n",c);
            break;
      }
   }

   motor_mask |= (FDC_INTS | FDC_RESET);

   floppy_start(0);
   for(l = 0; l < 1000000; l++);
   floppy_stop(0);

   /*
   printk("trying to seek...\n");
   floppy_start(0);
   floppy_seek();
   for(l = 0; l < 10000000; l++);
   floppy_stop(0);
   */
}

void floppy_isr(unsigned long irq)
{
}

#define NO_TRACK 255
static int reset=0;


static int result(void)
{
    int i = 0, counter, status;
    
    if (reset)
	return -1;
    for (counter = 0 ; counter < 10000 ; counter++) 
    {
	status = (int)in(FD_BASEIO+FD_STATUS)&(STATUS_DIR|STATUS_READY|STATUS_BUSY);
	if (status == STATUS_READY) 
	{
	    return i;
	}
	if (status == (STATUS_DIR|STATUS_READY|STATUS_BUSY)) 
	{
	    if (i >= MAX_REPLIES) 
	    {
		printk("le floppy a floode lors d'une reponse\n");
		break;
	    }
	    reply_buffer[i++] = in(FD_BASEIO+FD_DATA);
	}
    }
    reset = 1;
    current_track = NO_TRACK;
    printk("Result time-out\n");
    return -1;
}



